/*
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/
#include <sys/time.h>
#include <time.h>
#include <iostream>
#include <iomanip>

#include "slotsig.h"

double elapsed (timeval& tv1, timeval& tv2)
{
  double t1 = tv1.tv_sec + tv1.tv_usec / 1000000.0 ;
  double t2 = tv2.tv_sec + tv2.tv_usec / 1000000.0 ;
  return ((t2 - t1)*1000.0) ;
}

int main (int argc, char* argv[])
{
  unsigned nb = 1000 ;
  if ( argc > 1 )
    nb = atoi(argv[1]) ;
  if ( nb < 100 )
    nb = 1000 ;

  typedef nsSlotSig::clSig0<void> clSig ;
  typedef nsSlotSig::clSlot0<void> clSlot ;

  clSig**  sigs  = new (clSig*)[nb] ;
  clSlot** slots = new (clSlot*)[nb] ;
  unsigned cpt ;
  timeval tv1 ;
  timeval tv2 ;

  std::cout << "sizeof(nsSlotSig::clSig0<void>) = "
    << sizeof(nsSlotSig::clSig0<void>) << std::endl ;
  std::cout << "sizeof(nsSlotSig::clSlot0<void>) = "
    << sizeof(nsSlotSig::clSlot0<void>) << std::endl ;

  std::cout << "** Creating " << nb << " signals and slots...\n" ;
  gettimeofday(&tv1, 0) ;
  for (cpt = 0 ; cpt < nb ; ++cpt)
  {
    sigs[cpt] = new clSig ;
    slots[cpt] = new clSlot ;
  }
  gettimeofday(&tv2, 0) ;
  std::cout << "Time elapsed : " << std::setw(16) << elapsed(tv1, tv2) << "ms, "
    << std::setw(16) << std::dec
    << (unsigned long)((nb*2000.0)/elapsed(tv1, tv2)) << " units/sec.\n" ;

  std::cout << "** Connecting a signal to " << nb << " slots...\n" ;
  gettimeofday(&tv1, 0) ;
  for (cpt = 0 ; cpt < nb ; ++cpt)
    sigs[0]->connect(slots[cpt]) ;
  gettimeofday(&tv2, 0) ;
  std::cout << "Time elapsed : " << std::setw(16) << std::setw(16) << elapsed(tv1, tv2) << "ms, "
    << std::setw(16) << std::dec
    << (unsigned long)((nb*1000.0)/elapsed(tv1, tv2)) << " units/sec.\n" ;

  std::cout << "** Emitting the signal to " << nb << " slots...\n" ;
  gettimeofday(&tv1, 0) ;
  sigs[0]->emit() ;
  gettimeofday(&tv2, 0) ;
  std::cout << "Time elapsed : " << std::setw(16) << elapsed(tv1, tv2) << "ms, "
    << std::setw(16) << std::dec
    << (unsigned long)((nb*1000.0)/elapsed(tv1, tv2)) << " units/sec.\n" ;

  std::cout << "** Disconnecting the signal from " << nb << " slots...\n" ;
  gettimeofday(&tv1, 0) ;
  for (cpt = 0 ; cpt < nb ; ++cpt)
    sigs[0]->disconnect(slots[cpt]) ;
  gettimeofday(&tv2, 0) ;
  std::cout << "Time elapsed : " << std::setw(16) << elapsed(tv1, tv2) << "ms, "
    << std::setw(16) << std::dec
    << (unsigned long)((nb*1000.0)/elapsed(tv1, tv2)) << " units/sec.\n" ;

  std::cout << "** Connecting " << nb << " signals to " << nb << " slots...\n" ;
  gettimeofday(&tv1, 0) ;
  unsigned cpt2 ;
  for (cpt = 0 ; cpt < nb ; ++cpt)
    for (cpt2 = 0 ; cpt2 < nb ; ++cpt2)
      sigs[cpt2]->connect(slots[cpt]) ;
  gettimeofday(&tv2, 0) ;
  std::cout << "Time elapsed : " << std::setw(16) << elapsed(tv1, tv2) << "ms, "
    << std::setw(16) << std::dec
    << (unsigned long)((nb*1000.0*nb)/elapsed(tv1, tv2)) << " units/sec.\n" ;

  std::cout << "** Emitting " << nb << " signals to " << nb << " slots...\n" ;
  gettimeofday(&tv1, 0) ;
  for (cpt = 0 ; cpt < nb ; ++cpt)
      sigs[cpt]->emit() ;
  gettimeofday(&tv2, 0) ;
  std::cout << "Time elapsed : " << std::setw(16) << elapsed(tv1, tv2) << "ms, "
    << std::setw(16) << std::dec
    << (unsigned long)((nb*1000.0*nb)/elapsed(tv1, tv2)) << " units/sec.\n" ;

  std::cout << "** Deleting " << nb << " signals and slots...\n" ;
  gettimeofday(&tv1, 0) ;
  for (cpt = 0 ; cpt < nb ; ++cpt)
  {
    delete sigs[cpt] ;
    delete slots[cpt] ;
  }
  gettimeofday(&tv2, 0) ;
  std::cout << "Time elapsed : " << std::setw(16) << elapsed(tv1, tv2) << "ms, "
    << std::setw(16) << std::dec
    << (unsigned long)((nb*2000.0)/elapsed(tv1, tv2)) << " units/sec.\n" ;
  delete sigs ;
  delete slots ;

  return 0 ;
}
