/*
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#include <iostream>

#include "slotsig.h"

void testGlobalFunctions() ;
void testGlobalFunctionsAccum() ;
void testSlot() ;
void testSlotAccum() ;
void testSlotInClass() ;

int main (int argc, char* argv[])
{
  std::cout << "---------------------------------------------------------------\n" ;
  std::cout << "-- Testing SlotSig, return type = std::string, no parameters --\n" ;
  std::cout << "---------------------------------------------------------------\n" ;
  std::cout << "-- Please keep in mind that the actual order of calls might be\n"
            << "-- different than the given one in \"expected\" outputs.\n\n" ;

  testGlobalFunctions() ;
  testGlobalFunctionsAccum() ;
  testSlot() ;
  testSlotAccum() ;
  testSlotInClass() ;

  std::cout << "\n-------------------------\n"
            << "-- END OF TEST PROGRAM --\n"
            << "-------------------------\n" ;
  return 0 ;
}

//===========================================================================

std::string f0_0 ()
{
  std::cout << "f0_0()\n" ;
  return "1" ;
}

extern "C" std::string f1_0 ()
{
  std::cout << "f1_0()\n" ;
  return "2" ;
}

std::string f2_sender_0(nsSlotSig::clSenderBase* sender)
{
  std::cout << "f2_sender_0([" << std::hex << sender << "]), "
    << "sender name = " << sender->senderName() << "\n" ;
  return "3" ;
}

extern "C" std::string f3_sender_0(nsSlotSig::clSenderBase* sender)
{
  std::cout << "f3_sender_0([" << std::hex << sender << "]) " ;
  if ( sender != 0 )
    std::cout << ", sender name = " << sender->senderName() ;
  std::cout << "\n" ;
  return "4" ;
}

std::string GlobalString("") ;
void f_accum(const std::string& str)
{ GlobalString += std::string("/") + str ; }

class clAccum
{
  public:
    clAccum() : m_Str("") { }
    void operator() (const std::string& str_)
    { m_Str = std::string("/") + str_ + m_Str ; }
    const std::string& str() const
    { return m_Str ; }
    void reset()
    { m_Str = "" ; }
  private:
    std::string m_Str ;
} ;

class clSlot : public nsSlotSig::clSlot0<std::string>
{
  public:
    clSlot() : nsSlotSig::clSlot0<std::string>() { }
    std::string operator() ()
    {
      std::cout << "clSlot[" << std::hex << this << "]::operator(), sender = ["
        << std::hex << sender() << "]" ;
      if ( sender() != 0 )
        std::cout << ", sender name = " << sender()->senderName() ;
      std::cout << "\n" ;
      return "100" ;
    }
} ;

//===========================================================================

void testGlobalFunctions()
{
  std::cout << "\n******** Testing connections of a signal to global functions\n" ;

  std::cout << "* Declaring a signal, the return type being std::string : &sig_0 = " ;
  nsSlotSig::clSig0<std::string> sig_0 ;
  std::cout << std::hex << &sig_0 << std::endl ;
  std::cout << "* Creating a fake sender named \"MySender\", &sender = " ;
  nsSlotSig::clSenderBase sender("clSenderBase", "MySender") ;
  std::cout << std::hex << &sender << std::endl ;

  std::cout << "\n** Connecting a simple function, emitting, expected output :\n"
    << "* f0_0()\n" ;
  sig_0.connect(f0_0) ;
  sig_0.emit() ;

  std::cout << "\n** Connecting a simple extern \"C\" function, emitting, expected output :\n"
    << "* f0_0()\n"
    << "* f1_0()\n" ;
  sig_0.connect(f1_0) ;
  sig_0.emit() ;

  std::cout << "\n** Connecting a function receiving the sender, expected :\n"
    << "* f0_0()\n"
    << "* f1_0()\n"
    << "* f2_sender_0([" << std::hex << &sender << "]), sender name = " << sender.senderName() << "\n" ;
  sig_0.connect(f2_sender_0) ;
  sig_0.emit(&sender) ;

  std::cout << "\n** Connecting an extern \"C\" function receiving the sender, expected :\n"
    << "* f0_0()\n"
    << "* f1_0()\n"
    << "* f2_sender_0([" << std::hex << &sender << "]), sender name = " << sender.senderName() << "\n"
    << "* f3_sender_0([" << std::hex << &sender << "]), sender name = " << sender.senderName() << "\n" ;
  sig_0.connect(f3_sender_0) ;
  sig_0.emit(&sender) ;

  std::cout << "\n** Disconnecting the first and the last, expected :\n"
    << "* f1_0()\n"
    << "* f2_sender_0([" << std::hex << &sender << "]), sender name = " << sender.senderName() << "\n" ;
  sig_0.disconnect(f0_0) ;
  sig_0.disconnect(f3_sender_0) ;
  sig_0.emit(&sender) ;

  std::cout << "\n** Disconnecting all, expected nothing :\n" ;
  sig_0.disconnectAll() ;
  sig_0.emit() ;
}

void testGlobalFunctionsAccum()
{
  std::cout << "\n******** Testing connections of a signal to global functions, "
                "using an accumulator\n" ;

  std::cout << "* Declaring a signal, the return type being std::string : &sig_0 = " ;
  nsSlotSig::clSig0<std::string> sig_0 ;
  std::cout << std::hex << &sig_0 << std::endl ;
  std::cout << "* Creating a fake sender named \"MySender\", &sender = " ;
  nsSlotSig::clSenderBase sender("clSenderBase", "MySender") ;
  std::cout << std::hex << &sender << std::endl ;

  std::cout << "\n** Connecting a simple function, emitting using a function as accumulator...\n" ;
  sig_0.connect(f0_0) ;
  sig_0.emit(f_accum) ;
  std::cout << "* The accumulator result, expected  \"/1\" :\n" ;
  std::cout << GlobalString << std::endl ;

  std::cout << "\n** Resetting the accumulator to the empty string\n" ;
  GlobalString = "" ;
  std::cout << "** Connecting a function with sender, emitting using a function as accumulator...\n" ;
  sig_0.connect(f3_sender_0) ;
  sig_0.emit(f_accum, &sender) ;
  std::cout << "* The accumulator result, expected  \"/1/4\" :\n" ;
  std::cout << GlobalString << std::endl ;

  std::cout << "\n** Creating an instance of an accumulator class, &accum = " ;
  clAccum accum ;
  std::cout << std::hex << &accum << std::endl ;
  std::cout << "** Emitting, giving the accumulator instance, giving no sender...\n" ;
  sig_0.emit(accum) ;
  std::cout << "* The accumulator result, exepected \"/4/1\" : \n" ;
  std::cout << accum.str() << std::endl ;

}

void testSlot()
{
  std::cout << "\n******** Testing connections of a signal to slots\n" ;

  std::cout << "** Declaring a signal, the return type being std::string : &sig_0 = " ;
  nsSlotSig::clSig0<std::string> sig_0 ;
  std::cout << std::hex << &sig_0 << std::endl ;
  std::cout << "** Creating a fake sender named \"MySender\", &sender = " ;
  nsSlotSig::clSenderBase sender("clSenderBase", "MySender") ;
  std::cout << std::hex << &sender << std::endl ;

  std::cout << "\n** Reconnecting f0_0...\n" ;
  sig_0.connect(f0_0) ;
  std::cout << "** Reconnecting f3_sender_0...\n" ;
  sig_0.connect(f3_sender_0) ;

  std::cout << "** Creating two instance of a slot, &slot_0 = " ;
  clSlot slot_0 ;
  clSlot slot_1 ;
  std::cout << std::hex << &slot_0 << ", &slot_1 = "
            << std::hex << &slot_1 << std::endl ;
  std::cout << "** Connecting the signal to the two slots, emitting without sender, expected : \n" ;
  std::cout << "* clSlot[" << std::hex << &slot_0 << "]::operator(), sender = [0]\n" ;
  std::cout << "* clSlot[" << std::hex << &slot_1 << "]::operator(), sender = [0]\n" ;
  std::cout << "* f0_0()\n" ;
  std::cout << "* f3_sender_0([0])\n" ;
  sig_0.connect(&slot_0) ;
  sig_0.connect(&slot_1) ;
  sig_0.emit() ;

  std::cout << "\n** Disconnecting the slots from the signal, expected :\n" ;
  std::cout << "* f0_0()\n" ;
  std::cout << "* f3_sender_0([0])\n" ;
  sig_0.disconnect(&slot_0) ;
  sig_0.disconnect(&slot_1) ;
  sig_0.emit() ;

  std::cout << "\n** Connecting each slot to the signal, emitting with sender, expected : \n" ;
  std::cout << "* clSlot[" << std::hex << &slot_0 << "]::operator(), sender = [" << std::hex << &sender << "]\n" ;
  std::cout << "* clSlot[" << std::hex << &slot_1 << "]::operator(), sender = [" << std::hex << &sender << "]\n" ;
  std::cout << "* f0_0()\n" ;
  std::cout << "* f3_sender_0([" << std::hex << &sender << "])\n" ;
  slot_0.connect(&sig_0) ;
  slot_1.connect(&sig_0) ;
  sig_0.emit(&sender) ;

  std::cout << "\n** Calling slot_1.disconnectAll(), expected :\n" ;
  std::cout << "* clSlot[" << std::hex << &slot_0 << "]::operator(), sender = [" << std::hex << &sender << "]\n" ;
  std::cout << "* f0_0()\n" ;
  std::cout << "* f3_sender_0([" << std::hex << &sender << "])\n" ;
  slot_1.disconnectAll() ;
  sig_0.emit(&sender) ;

  std::cout << "\n** Calling sig_0.disconnectAll(), expected nothing.\n" ;
  sig_0.disconnectAll() ;
  sig_0.emit() ;
}

void testSlotAccum()
{
  std::cout << "\n******** Testing connections of a signal to slots, "
               "using an accumulator\n" ;

  std::cout << "** Declaring a signal, the return type being std::string : &sig_0 = " ;
  nsSlotSig::clSig0<std::string> sig_0 ;
  std::cout << std::hex << &sig_0 << std::endl ;

  std::cout << "** Connecting the same two previous functions and the two slots...\n" ;
  clSlot slot_0 ;
  clSlot slot_1 ;
  sig_0.connect(f0_0) ;
  sig_0.connect(f3_sender_0) ;
  sig_0.connect(&slot_1) ;
  sig_0.connect(&slot_0) ;

  std::cout << "** Using the global function as accumulator, expected result :\n"
            << "* /100/100/1/4\n" ;
  GlobalString = "" ;
  sig_0.emit(f_accum) ;
  std::cout << GlobalString << std::endl ;

  clAccum accum ;
  std::cout << "** Using the global function as accumulator, expected result :\n"
            << "* /4/1/100/100\n" ;
  sig_0.emit(accum) ;
  std::cout << accum.str() << std::endl ;

}

class cl1 : public nsSlotSig::clSenderBase
{
  public:
    cl1() :
      nsSlotSig::clSenderBase("class cl1", "cl1 instance"),
      m_Sig(this),
      m_Slot(),
      m_SlotInClass(this, &cl1::someMethod)
    { }

    typedef nsSlotSig::clSig0<void> clSig_cl1 ;
    typedef nsSlotSig::clSlotInClass0<cl1, void> clSlotInClass_cl1 ;

    class clSlot_cl1 : public nsSlotSig::clSlot0<void>
    {
      public:
        clSlot_cl1() : nsSlotSig::clSlot0<void>()
        { }
        virtual void operator() ()
        { std::cout << "cl1::clSlot_cl1::operator(), " ;
          if ( sender() != 0 )
            std::cout << "sender name = " << sender()->senderName() << std::endl ;
          else
            std::cout << "no sender !\n" ;
        }
    } ;

    clSig_cl1* sig()
    { return &m_Sig ; }

    clSlot_cl1* slot()
    { return &m_Slot ; }

    clSlotInClass_cl1* slotInClass()
    { return &m_SlotInClass ; }

    void someMethod()
    {
      std::cout << "cl1::someMethod(), " ;
      if ( m_SlotInClass.sender() != 0 )
        std::cout << "sender name = " << m_SlotInClass.sender()->senderName() ;
      else
        std::cout << "no sender !" ;
      std::cout << std::endl ;
    }

  private:
    clSig_cl1         m_Sig ;
    clSlot_cl1        m_Slot ;
    clSlotInClass_cl1 m_SlotInClass ;
} ;

class cl2 : public nsSlotSig::clSenderBase
{
  public:
    cl2() :
      nsSlotSig::clSenderBase("class cl2", "cl2 instance"),
      m_Sig(this),
      m_Slot(),
      m_SlotInClass(this, &cl2::someOtherMethod)
    { }

    typedef nsSlotSig::clSig0<void> clSig_cl2 ;
    typedef nsSlotSig::clSlotInClass0<cl2, void> clSlotInClass_cl2 ;

    class clSlot_cl2 : public nsSlotSig::clSlot0<void>
    {
      public:
        void operator() ()
        { std::cout << "cl2::clSlot_cl2::operator(), " ;
          if ( sender() != 0 )
            std::cout << "sender name = " << sender()->senderName() << std::endl ;
          else
            std::cout << "no sender !\n" ;
        }
    } ;

    clSig_cl2* sig()
    { return &m_Sig ; }

    clSlot_cl2* slot()
    { return &m_Slot ; }

    clSlotInClass_cl2* slotInClass()
    { return &m_SlotInClass ; }

    void someOtherMethod()
    {
      std::cout << "cl2::someOtherMethod(), " ;
      if ( m_SlotInClass.sender() != 0 )
        std::cout << "sender name = " << m_SlotInClass.sender()->senderName() ;
      else
        std::cout << "no sender !" ;
      std::cout << std::endl ;
    }

  private:
    clSig_cl2         m_Sig ;
    clSlot_cl2        m_Slot ;
    clSlotInClass_cl2 m_SlotInClass ;
} ;

void testSlotInClass()
{
  std::cout << "\n******** Testing connections of signals and slots in classes\n" ;

  std::cout << "** Creating a signal returning void...\n" ;
  nsSlotSig::clSig0<void> sig_0 ;
  std::cout << "** Creating an instance of cl1 and cl2...\n" ;
  cl1 c1 ;
  cl2 c2 ;

  std::cout << "\n** Connecting the signal to each slot of cl1 and cl2...\n" ;

  sig_0.connect(c1.slot()) ;
  sig_0.connect(c1.slotInClass()) ;
  sig_0.connect(c2.slot()) ;
  sig_0.connect(c2.slotInClass()) ;
  std::cout << "** Emitting, expected :\n"
    << "* cl1::clSlot_cl1::operator(), no sender !\n"
    << "* cl1::someMethod(), no sender !\n"
    << "* cl2::clSlot_cl2::operator(), no sender !\n"
    << "* cl2::someOtherMethod(), no sender !\n" ;
  sig_0.emit() ;

  std::cout << "\n** Connecting the signal from cl1 to the slots of cl2...\n" ;
  c1.sig()->connect(c2.slot()) ;
  c1.sig()->connect(c2.slotInClass()) ;
  std::cout << "\n** Emitting signal from cl1, expected :\n"
    << "* cl2::clSlot_cl2::operator(), sender name = cl1 instance\n"
    << "* cl2::someOtherMethod(), sender name = cl1 instance\n" ;
  c1.sig()->emit() ;

  std::cout << "\n** Connecting the slots of cl1 to the signal of cl2...\n" ;
  c1.slot()->connect(c2.sig()) ;
  c1.slotInClass()->connect(c2.sig()) ;
  std::cout << "\n** Emitting signal from cl2, expected :\n"
    << "* cl1::clSlot_cl1::operator(), sender name = cl2 instance\n"
    << "* cl1::someOtherMethod(), sender name = cl2 instance\n" ;
  c2.sig()->emit() ;

}
